	page	60,132
;-----------------------------------------------------------------------------
;
; UNREAL.ASM
;
;	Copyright (c) 1998-Present  Robert Collins
;
;       You have my permission to copy and distribute this software for
;       non-commercial purposes.  Any commercial use of this software or
;       source code is allowed, so long as the appropriate copyright
;       attributions (to me) are intact, *AND* my email address is properly
;       displayed.
;
;       Basically, give me credit, where credit is due, and show my email
;       address.
;
;-----------------------------------------------------------------------------
;
;       Robert R. Collins               email:  rcollins@x86.org
;
;-----------------------------------------------------------------------------


;-----------------------------------------------------------------------------
; Assembler directives
;-----------------------------------------------------------------------------
	.xlist			; disable list file
	.586P


;-----------------------------------------------------------------------------
; Structures
;-----------------------------------------------------------------------------
    Descriptor	STRUC
	Seg_limit	dw	?	; Segment limit
	Base_A15_A00	dw	?	; A00..A15 of base address
	Base_A23_A16	db	?	; A16..A23 of base address
	Access_rights	db	?	; Segment access rights
	GDLimit_A19_A16 db	?	; Granularity, Op-size, Limit A16..A19
	Base_A31_A24	db	?	; A24..A31 of base address
    Descriptor	ENDS


;-----------------------------------------------------------------------------
; Equates & macros
;-----------------------------------------------------------------------------
	CRLF$		equ	0dh,0ah,24h


JMPFAR	MACRO	destination,selector	; dynamic JMP FAR SEG:OFF
	db	0eah			;; jmp instruction
	dw	offset destination	;; offset word
	dw	selector		;; segment selector word
	endm


	.list
	_DATA segment para public use16 'DATA'
;-----------------------------------------------------------------------------
; Data segment
;-----------------------------------------------------------------------------
	GDT_386 label	fword
	GDT_PTR 	Descriptor	<>
	SEL_CS4G	equ	$-GDT_386
	GDT_CS4G	Descriptor	<-1,,,9bh,8fh,> ; CS Descriptor
	SEL_DS4G	equ		$-GDT_386
	GDT_DS4G	Descriptor	<-1h,0,0h,93h,8fh,0h>	; 4G Descriptor
	SEL_STACK4G	equ		$-GDT_386
	GDT_STACK4G	Descriptor	<-1h,0,0h,93h,8fh,0h>	; 4G Descriptor
	GDT_Len 	equ	($-GDT_386) - 1


;-----------------------------------------------------------------------------
; All other data
;-----------------------------------------------------------------------------
	PG_Msg		db	"Can't run test...already in paging mode.",CRLF$
	PE_Msg		db	"Can't run test...already in protected mode.",CRLF$
	Passed_Msg	db	"Unreal mode now enabled.",CRLF$

	_DATA	ENDS

	_TEXT	segment para public use16 'CODE'
	ASSUME	CS:_TEXT, DS:_DATA, ES:_DATA, SS:STACK

;-----------------------------------------------------------------------------
; Code starts here
;-----------------------------------------------------------------------------
	_UNREAL        proc    far
	mov	ax,seg STACK		; setup stack segment
	mov	ss,ax
	mov	sp,sizeof StackPtr
	xor	ax,ax			; clear it
	pushf
	push	ds			; save far return on stack
	push	ax

;-----------------------------------------------------------------------------
; Set segments to normal data segment
;-----------------------------------------------------------------------------
@@:	mov	ax,seg _DATA		; get original data segment
	mov	ds,ax
	mov	es,ax

;-----------------------------------------------------------------------------
; Check that this processor supports 4M pages.
;-----------------------------------------------------------------------------
	xor	eax,eax 		; clear it
	db	66h			; were're going to do something
	smsw	ax			;  undocumented here, don't look
	test	eax,80000001h		; page mode or pretected mode enabled?
	jz	@F			; nope, continue
	mov	dx,offset PG_Msg	;
	shl	eax,1			; check for PE
	jc	@ErrorExit		; oops
	mov	dx,offset PE_Msg

@ErrorExit:
	mov	ah,9
	int	21h			; print message
	mov	ax,4c01h		; set error code
	int	21h
	iret				; go split, just in case

;-----------------------------------------------------------------------------
; Setup descriptor table
;-----------------------------------------------------------------------------
@@:	mov	eax,ds			; make pointer to GDT table
	shl	eax,4			; have physical address of segment
	add	eax,offset GDT_386	; now have physical addr of table
	mov	GDT_PTR.Seg_limit,GDT_Len	; set length
	mov	GDT_PTR.Base_A15_A00,ax
	shr	eax,10h 		; get other address bits
	mov	GDT_PTR.Base_A23_A16,al
	mov	GDT_PTR.Access_rights,ah

	mov	eax,cs			; get CS
	shl	eax,4			; now have physical address
	mov	GDT_CS4G.Base_A15_A00,ax
	shr	eax,10h 		; get other address bits
	mov	GDT_CS4G.Base_A23_A16,al
	mov	GDT_CS4G.Base_A31_A24,ah

	mov	eax,ds			; get DS
	shl	eax,4			; now have physical address
	mov	GDT_DS4G.Base_A15_A00,ax
	shr	eax,10h 		; get other address bits
	mov	GDT_DS4G.Base_A23_A16,al
	mov	GDT_DS4G.Base_A31_A24,ah

	mov	eax,ss			; get DS
	shl	eax,4			; now have physical address
	mov	GDT_STACK4G.Base_A15_A00,ax
	shr	eax,10h 		; get other address bits
	mov	GDT_STACK4G.Base_A23_A16,al
	mov	GDT_STACK4G.Base_A31_A24,ah

;-----------------------------------------------------------------------------
; Initialize page mode
;-----------------------------------------------------------------------------
	cli
	lgdt	GDT_386
	mov	eax,cr0 		; get control register
	or	al,1			;
	mov	cr0,eax
	push	cs			; push return selector on stack
	push	offset PMRET		; set return offset
	JMPFAR	@F,SEL_CS4G

@@:	mov	ax,SEL_DS4G		; set all selectors to 4GB limits
	mov	ds,ax
	mov	es,ax
	mov	fs,ax
	mov	gs,ax
	mov	ax,SEL_STACK4G
	mov	ss,ax

;-----------------------------------------------------------------------------
; Split from this program
;-----------------------------------------------------------------------------
	mov	ebx,cr0 		; get 386 control register
	and	ebx,not 80000001h	; clear paging bit
	mov	cr0,ebx 		;   and store in CR0
	retf

PMRET:
	mov	ax,seg _DATA
	mov	ds,ax
	mov	es,ax
	mov	fs,ax
	mov	gs,ax
	mov	ax,seg STACK
	mov	ss,ax
	mov	dx,offset Passed_Msg
	mov	ah,9
	int	21h			; print message
	mov	ax,4c00h		; set error code
	int	21h
	iret				; go split, just in case
	_UNREAL        endp

_TEXT	ends


	STACK segment para public 'STACK'
;-----------------------------------------------------------------------------
; Stack segment
;-----------------------------------------------------------------------------
	StackPtr	db	400h dup (?)
	STACK	ends

	end	_UNREAL
